import pytest

from mhdwriter.args import WriteType
from mhdwriter.header import generate_header, WriteArgs


@pytest.fixture
def basic_args():
    args = WriteArgs(
        write_type=WriteType.RAW,
        fov="C",
        series_description="test series",
        study_description="test study",
        protocol="test protocol",
        date_time="2021-01-01 00:00:00",
        root_site_id="12345567",
        downsample_factor=0,
        height=0,
        width=0,
        length=0,
        is_rgb=False,
    )

    yield args


def test_generate_header_should_require_args():
    with pytest.raises(AssertionError) as excinfo:
        generate_header(1)
    assert "Invalid WriteArgs" in str(excinfo.value)


def test_generate_header_should_generate_uuids(basic_args):
    header = generate_header(basic_args)
    assert header["StudyInstanceUID"] == "12345567.7189148.32749.210101.0.1"
    assert header["SeriesInstanceUID"] == "12345567.7189148.8134279.210101.0.2"
    assert header["StudyInstanceUID"] == "12345567.7189148.32749.210101.0.1"


def test_generate_header_should_include_all_fields(basic_args):
    header = generate_header(basic_args)
    assert len(header) == 28


def test_generate_header_should_include_pixel_size_by_fov(basic_args):
    header = generate_header(basic_args)
    assert header["ElementSpacing"] == "0.035 0.035 0.035"
    elspacing1 = header["ElementSpacing"]
    basic_args.fov = "A"
    header = generate_header(basic_args)
    assert header["ElementSpacing"] != elspacing1
    assert header["ElementSpacing"] == "0.02 0.02 0.02"


def test_generate_header_should_include_downsampled_pixel_size(basic_args):
    basic_args.downsample_factor = 1
    header = generate_header(basic_args)
    assert header["ElementSpacing"] == "0.07 0.07 0.07"
    basic_args.downsample_factor = 2
    header = generate_header(basic_args)
    assert header["ElementSpacing"] == "0.14 0.14 0.14"



import shutil
from pathlib import Path
from typing import Optional

import cv2
import numpy as np

from mhdwriter.args import WriteArgs, WriteType
from mhdwriter.header import generate_header


def write_mhd_raw(
    input_dir: Path, args: WriteArgs, out_path: Optional[Path] = None
) -> bool:
    """
    Convert a stack of files in a directory to a mhd/raw pair.

    Args:
        input_dir (Path): The directory containing the stack of files to convert to a volume.
        args (WriteArgs): WriteArgs object specifying parameters for the MHD header creation.
        out_path (Path): Optional path to specify output directory/file_base. If missing, metadata will be used
            to determine output file/path name in the input_dir.

    Returns:
        bool: Returns True upon successful mhd/raw creation, else False.

    """
    all_files = sorted(list(input_dir.glob("*")))
    all_files = [
        f
        for f in all_files
        if f.is_file() and f.suffix.lower()
        in [".jpg", ".jxl", ".jp2", ".jpeg", ".png", ".tif", ".tiff"]
    ]
    img = cv2.imread(str(all_files[0]), cv2.IMREAD_UNCHANGED)
    args.height, args.width = img.shape[:2]
    if args.downsample_factor > 0:
        dx_factor = 2 ** args.downsample_factor
        args.height = args.height//dx_factor
        args.width = args.width//dx_factor
    args.length = len(all_files)
    metadata = generate_header(args)

    if not input_dir.exists():
        raise FileNotFoundError(f"Missing input dir {input_dir}")

    if out_path is None:
        out_path = input_dir

    if args.write_type == WriteType.NONE:
        out_path = out_path.joinpath(f"{metadata['SeriesDescription']}")
        out_path.mkdir(exist_ok=True, parents=True)
    else:
        out_path = out_path.joinpath(f"{metadata['ElementDataFile']}")

    if out_path.exists():
        print(f"Output path {out_path} already exists")
        return

    with out_path.open(mode="wb") as outfile:
        for idx, img_file in enumerate(all_files):
            try:
                img = cv2.imread(str(img_file), cv2.IMREAD_UNCHANGED)
            except Exception as e:
                print(f"Error reading {str(img_file)}: {e}")
                continue
            if img is None:
                print(f"Error reading {str(img_file)}")
                continue
            if args.is_rgb:
                img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

            if args.downsample_factor > 0:
                img = cv2.resize(
                    img, (args.width, args.height), interpolation=cv2.INTER_AREA
                )

            if args.write_type == WriteType.NONE:
                file_path = out_path.joinpath("slice_{0:04d}.jpg".format(idx + 1))
                cv2.imwrite(str(file_path), img)
            else:
                squeezed = np.squeeze(img)
                outfile.write(squeezed.tobytes())

    if not out_path.exists():
        return False
    mhd_path = out_path.parent.joinpath(f"{out_path.stem}.mhd")
    with mhd_path.open("w") as mhd:
        for key in metadata:
            mhd.write(f"{key} = {metadata[key]}\n")
    return True



import hashlib
from datetime import datetime
from typing import Any, Dict

from mhdwriter.args import WriteArgs, WriteType


def generate_header(args: WriteArgs) -> Dict[str, Any]:
    """
    Generate a .mhd file for a mhd/raw pair.

    Args:
        args (WriteArgs): WriteArgs object specifying parameters for the MHD header creation.

    Returns:
        dict: Returns an ordrered dictionary of relevant metadata key->value pairs.

    """

    assert type(args) == WriteArgs, "Invalid WriteArgs"

    if isinstance(args.date_time, str):
        study_date_time = parse_datetime(args.date_time)
    else:
        study_date_time = args.date_time

    is_multi = True if args.write_type == WriteType.NONE else False

    raw_name = args.series_description + ".raw"
    raw_name = raw_name.replace(" ", "_")
    if args.write_type == WriteType.RAW_COMPRESSED:
        raw_name += ".gz"
    pixel_sz = _pixel_size(args.fov.upper()) * (2 ** args.downsample_factor)
    pixel_sz = round(pixel_sz, 5)

    metadata = {
        "ObjectType": "Image",
        "NDims": 3,
        "BinaryData": True,
        "BinaryDataByteOrderMSB": False,
        "Modality": "CFT",
        "ManufacturersModelName": "Xerra",
        "TransformMatrix": "-1 0 0 0 -1 0 0 0 -1",
        "Offset": f"{round(pixel_sz * args.width, 3)} {round(pixel_sz * args.height, 3)} {round(float(pixel_sz) * args.length, 3)}",
        "CenterOfRotation": "0 0 0",
        "AnatomicalOrientation": "RAI",
        "ElementSpacing": f"{pixel_sz} {pixel_sz} {pixel_sz}",
        "DimSize": f"{args.width} {args.height} {args.length}",
        "FOV": args.fov,
        "PatientsName": args.study_description,
        "PatientID": args.study_description,
        "StudyID": args.study_description,
        "StudyDescription": args.protocol,
        "SeriesDescription": args.series_description,
        "Protocol": args.protocol,
        "StudyDate": study_date_time.strftime("%Y%m%d"),
        "StudyTime": study_date_time.strftime("%H%M%S"),
    }

    metadata = _setup_uids(metadata, args.root_site_id)
    if args.is_rgb:
        metadata["ShowRGB"] = 1
        if not args.write_type == WriteType.NONE:
            metadata["ElementNumberOfChannels"] = 3
        metadata["ElementType"] = "MET_UCHAR"
    elif args.is_hdr == "float32":
        metadata["ElementType"] = "MET_FLOAT"
    else:
        metadata["ElementType"] = "MET_USHORT"

    if is_multi:
        metadata["ElementDataFile"] = "slice_%04d.jpg 0 " + str(args.length) + " 1"
    else:
        metadata["ElementDataFile"] = raw_name
    return metadata


def _pixel_size(fov: str) -> float:
    fov = fov.upper()
    xy = 0.00
    if fov == "A":
        xy = 0.020
    elif fov == "B":
        xy = 0.025
    elif fov == "C":
        xy = 0.035
    elif fov == "D":
        xy = 0.045
    elif fov == "E":
        xy = 0.055
    return xy


def _setup_uids(metadata_dict: dict[str, Any], root_site_id: str) -> Dict:
    patients_name = metadata_dict.get("PatientsName", "")
    patients_id = metadata_dict.get("PatientID", "")
    study_id = metadata_dict.get("StudyID", "")
    study_date = int(metadata_dict.get("StudyDate", 20000000))
    study_time = int(metadata_dict.get("StudyTime", 0))
    study_description = metadata_dict.get("StudyDescription", "")
    series_description = metadata_dict.get("SeriesDescription", study_description)
    hash_str = f"{patients_name}{patients_id}{study_id}{study_description}"
    m = hashlib.md5(hash_str.encode()).hexdigest()
    id0 = int(m[0:6], base=16)
    id1 = int(m[-4:], base=16)
    uid_series = hashlib.md5(f"{m}{series_description}".encode()).hexdigest()
    uid_series_int = int(uid_series[0:6], base=16)
    uid_image = int(hashlib.md5(f"{uid_series}1".encode()).hexdigest()[0:6], base=16)
    metadata_dict[
        "StudyInstanceUID"
    ] = f"{root_site_id}.{id0}.{id1}.{study_date - 20000000}.{study_time}.1"
    metadata_dict[
        "SeriesInstanceUID"
    ] = f"{root_site_id}.{id0}.{uid_series_int}.{study_date - 20000000}.{study_time}.2"
    metadata_dict[
        "SOPInstanceUID"
    ] = f"{root_site_id}.{id0}.{uid_image}.{study_date - 20000000}.{study_time}.3"
    metadata_dict["SeriesDate"] = metadata_dict["StudyDate"]
    metadata_dict["SeriesTime"] = metadata_dict["StudyTime"]
    return metadata_dict


def parse_datetime(date_string) -> datetime:
    formats = ["%Y-%m-%d-%H:%M:%S", "%Y-%m-%dT%H:%M:%S", "%Y-%m-%d %H:%M:%S"]
    for fmt in formats:
        try:
            return datetime.strptime(date_string, fmt)
        except ValueError:
            continue
    raise ValueError(f"time data '{date_string}' does not match expected formats")



from dataclasses import dataclass
from datetime import datetime
from enum import Enum


class WriteType(Enum):
    NONE = 0
    RAW = 1
    RAW_COMPRESSED = 2


@dataclass
class WriteArgs:
    fov: str
    series_description: str
    study_description: str
    protocol: str
    date_time: datetime | str
    write_type: WriteType = WriteType.RAW
    root_site_id: str = "1.2.826.0.1.3680043.10.1341"
    downsample_factor: int = 0
    height: int = 0
    width: int = 0
    length: int = 0
    is_rgb: bool = False
    is_hdr: bool = False



import shutil
from pathlib import Path

import pytest

from mhdwriter.writer import write_mhd_raw
from .test_header import basic_args


@pytest.fixture
def test_jpg_dataset(tmp_path):
    test_data_dir = get_test_data_dir()
    for i in range(1, 30):
        source_file = test_data_dir / "example.jpg"
        shutil.copy(source_file, str(Path(tmp_path).joinpath(f"slice_{i:04d}.jpg")))
    yield tmp_path


@pytest.fixture
def test_tiff_dataset(tmp_path):
    test_data_dir = get_test_data_dir()
    for i in range(1, 60):
        source_file = test_data_dir / "example.tif"
        shutil.copy(source_file, str(Path(tmp_path).joinpath(f"slice_{i:04d}.tiff")))

    yield tmp_path


def get_test_data_dir():
    current_file_path = Path(__file__).resolve().parent.parent
    return current_file_path / "test_data"


def test_write_mhd_jpg_raw(test_jpg_dataset, basic_args):
    basic_args.is_rgb = True
    write_mhd_raw(test_jpg_dataset, basic_args)
    assert True


def test_write_mhd_tiff_raw(test_tiff_dataset, basic_args):
    basic_args.is_rgb = False
    write_mhd_raw(test_tiff_dataset, basic_args)
    assert True
